from typing import Optional

DLT_PIPELINE_COMMAND_DOCS_URL = (
    "https://dlthub.com/docs/reference/command-line-interface#dlt-pipeline"
)
DLT_INIT_DOCS_URL = "https://dlthub.com/docs/reference/command-line-interface#dlt-init"
DLT_TELEMETRY_DOCS_URL = "https://dlthub.com/docs/reference/telemetry"
DLT_DEPLOY_DOCS_URL = "https://dlthub.com/docs/walkthroughs/deploy-a-pipeline"
DLT_AI_DOCS_URL = "https://dlthub.com/docs/dlt-ecosystem/llm-tooling/llm-native-workflow"

DEFAULT_AI_WORKBENCH_REPO = "https://github.com/dlt-hub/dlthub-ai-workbench.git"
DEFAULT_AI_WORKBENCH_BRANCH: Optional[str] = None  # must be None for dlt release!
DEFAULT_AI_WORKBENCH_LICENSE_URL = (
    "https://github.com/dlt-hub/dlthub-ai-workbench/blob/master/LICENSE"
)
